""" Utilities for tests """

from bodhi.messages.schemas.update import UpdateRequestTestingV1


def get_rule(rules, name):
    for rule in rules:
        if rule["name"] == name:
            return rule


example_real_bodhi_message = UpdateRequestTestingV1(
    topic="org.fedoraproject.prod.bodhi.update.request.testing",
    body={
        "agent": "lmacken",
        "update": {
            "alias": "FEDORA-2019-f1ca3c00e5",
            "status": "pending",
            "critpath": False,
            "stable_karma": 3,
            "date_pushed": None,
            "title": "gnome-settings-daemon-3.6.1-1.fc18,control-center-3.6.1-1.fc18",
            "nagged": None,
            "comments": [
                {
                    "group": None,
                    "author": "bodhi",
                    "text": "This update has been submitted for testing by hadess. ",
                    "karma": 0,
                    "anonymous": False,
                    "timestamp": 1349718539.0,
                    "update_title": "gnome-settings-daemon-3.6.1-1.fc18,"
                    + "control-center-3.6.1-1.fc18",
                }
            ],
            "updateid": None,
            "type": "bugfix",
            "close_bugs": True,
            "date_submitted": 1349718534.0,
            "unstable_karma": -3,
            "release": {
                "dist_tag": "f18",
                "locked": True,
                "long_name": "Fedora 18",
                "name": "F18",
                "id_prefix": "FEDORA",
            },
            "approved": None,
            "builds": [
                {
                    "nvr": "gnome-settings-daemon-3.6.1-1.fc18",
                    "package": {
                        "suggest_reboot": False,
                        "committers": ["hadess", "ofourdan", "mkasik", "cosimoc"],
                        "name": "gnome-settings-daemon",
                    },
                },
                {
                    "nvr": "control-center-3.6.1-1.fc18",
                    "package": {
                        "suggest_reboot": False,
                        "committers": [
                            "ctrl-center-team",
                            "ofourdan",
                            "ssp",
                            "ajax",
                            "alexl",
                            "jrb",
                            "mbarnes",
                            "caolanm",
                            "davidz",
                            "mclasen",
                            "rhughes",
                            "hadess",
                            "johnp",
                            "caillon",
                            "whot",
                            "rstrode",
                        ],
                        "name": "control-center",
                    },
                },
            ],
            "date_modified": None,
            "notes": "This update fixes numerous bugs in the new Input "
            + "Sources support, the Network panel and adds a help "
            + "screen accessible via Wacom tablets's buttons.",
            "request": "testing",
            "bugs": [],
            "critpath_approved": False,
            "karma": 0,
            "submitter": "hadess",
            "user": {"name": "hadess"},
        },
    },
)



import logging
from unittest.mock import Mock, patch

import pytest
from fedora_messaging.message import Message

from .utils import get_rule


@pytest.fixture
def rule(rules):
    return get_rule(rules, "Junior Tagger (Tagger I)")


@pytest.fixture
def message():
    return Message(
        topic="org.fedoraproject.prod.fedoratagger.tag.create",
    )


def test_complicated_trigger_against_empty(rule, message):
    assert rule.matches(message, Mock(name="tahrir")) == set()


def test_complicated_trigger_against_partial(rule, message):
    message.body = {"user": {}}
    assert rule.matches(message, Mock(name="tahrir")) == set()


def test_complicated_trigger_against_partial_mismatch(rule, message, tahrir_client, caplog):
    caplog.set_level(logging.ERROR)
    message.body = {"user": None}
    assert rule.matches(message, tahrir_client) == set()
    print(caplog.messages)
    print(caplog.text)
    assert len(caplog.messages) == 1


def test_complicated_trigger_against_full_match(rule, message, tahrir_client):
    message.body = {
        "tag": {
            "dislike": 0,
            "like": 1,
            "package": "mattd",
            "tag": "awesome",
            "total": 1,
            "votes": 1,
        },
        "user": {"anonymous": False, "rank": -1, "username": "ralph", "votes": 4},
        "vote": {
            "like": True,
            "tag": {
                "dislike": 0,
                "like": 1,
                "package": "mattd",
                "tag": "awesome",
                "total": 1,
                "votes": 1,
            },
            "user": {"anonymous": False, "rank": -1, "username": "ralph", "votes": 4},
        },
    }

    # Set up some mock stuff
    class MockQuery:
        def count(self):
            return float("inf")  # Master tagger

    with (
        patch("fedbadges.rules.user_exists_in_fas") as g,
        patch("datanommer.models.Message.grep") as grep,
    ):
        grep.return_value = float("inf"), 1, MockQuery()
        g.return_value = True
        assert rule.matches(message, tahrir_client) == {"ralph"}



import inspect
from unittest.mock import Mock, patch

import datanommer
import pytest
from fedora_messaging.config import conf
from tahrir_api import dbapi

from fedbadges.consumer import FedoraBadgesConsumer
from fedbadges.rulesrepo import RulesRepo


@pytest.fixture()
def grep():
    grep_spec = inspect.getfullargspec(datanommer.models.Message.grep)
    print(grep_spec)
    with patch("datanommer.models.Message.grep", spec=grep_spec.args) as patched_grep:
        yield patched_grep


@pytest.fixture()
def fm_config(tmp_path):
    test_config = dict(
        badges_repo="tests/test_badges",
        database_uri=f"sqlite:///{tmp_path.as_posix()}/badges.db",
        datanommer_db_uri=f"sqlite:///{tmp_path.as_posix()}/datanommer.db",
        datagrepper_url="https://example.com/datagrepper",
        distgit_hostname="src.example.com",
        id_provider_hostname="id.example.com",
        fasjson_base_url="https://fasjson.example.com",
        badge_issuer=dict(
            issuer_id="test-issuer",
            issuer_name="Testing",
            issuer_origin="http://badges.example.com",
            issuer_url="http://example.com",
            issuer_email="badges@example.com",
        ),
    )
    with patch.dict(conf["consumer_config"], test_config):
        yield


@pytest.fixture()
def notification_callback_mock():
    return Mock(name="notification_callack")


@pytest.fixture()
def badges_db(tmp_path, notification_callback_mock):
    database_uri = f"sqlite:///{tmp_path.as_posix()}/badges.db"
    db = dbapi.TahrirDatabase(
        dburi=database_uri,
        autocommit=False,
        notification_callback=notification_callback_mock,
    )
    db.db_mgr.sync()
    yield db
    db.session.rollback()


@pytest.fixture()
def fasjson_client():
    client = Mock(name="fasjson")
    with patch("fedbadges.consumer.fasjson_client.Client", return_value=client):
        yield client


@pytest.fixture()
def consumer(fm_config, badges_db, fasjson_client):
    return FedoraBadgesConsumer()


@pytest.fixture()
def tahrir_client(fm_config, badges_db, notification_callback_mock):
    issuer = conf["consumer_config"]["badge_issuer"]
    badges_db.add_issuer(
        issuer.get("issuer_origin"),
        issuer.get("issuer_name"),
        issuer.get("issuer_url"),
        issuer.get("issuer_email"),
    )
    badges_db.session.commit()
    yield badges_db


@pytest.fixture()
def rules(fm_config, fasjson_client, tahrir_client):
    repo = RulesRepo(conf["consumer_config"], 1, fasjson_client)
    return repo.load_all(tahrir_client=tahrir_client)



from unittest.mock import patch

import pytest
from fedora_messaging.message import Message

from .utils import get_rule


class MockQuery:
    def count(self):
        return float("inf")  # Master tagger


def test_complicated_recipient_real(
    rules,
    tahrir_client,
):
    rule = get_rule(rules, "Speak Up!")
    msg = Message(
        topic="org.fedoraproject.prod.meetbot.meeting.complete",
        body={
            "meeting_topic": "testing",
            "attendees": {"zodbot": 2, "threebean": 2},
            "chairs": {},
            "topic": "",
            "url": "fedora-meeting.2013-06-24-19.52",
            "owner": "threebean",
            "channel": "#fedora-meeting",
        },
    )
    with (
        patch("fedbadges.rules.user_exists_in_fas") as g,
        patch("datanommer.models.Message.grep") as grep,
    ):
        grep.return_value = float("inf"), 1, MockQuery()
        g.return_value = True
        assert rule.matches(msg, tahrir_client) == {"zodbot", "threebean"}


def test_complicated_recipient_pagure(
    rules,
    tahrir_client,
):
    rule = get_rule(rules, "Long Life to Pagure (Pagure I)")
    msg = Message(
        topic="io.pagure.prod.pagure.git.receive",
        body={
            "authors": [
                {"fullname": "Pierre-YvesChibon", "name": "pingou"},
                {"fullname": "Lubom\\u00edr Sedl\\u00e1\\u0159", "name": "lsedlar"},
            ],
            "total_commits": 2,
            "start_commit": "da090b8449237e3878d4d1fe56f7f8fcfd13a248",
        },
    )

    with (
        patch("fedbadges.rules.user_exists_in_fas") as g,
        patch("datanommer.models.Message.grep") as grep,
    ):
        grep.return_value = float("inf"), 1, MockQuery()
        g.return_value = True
        assert rule.matches(msg, tahrir_client) == {"pingou", "lsedlar"}


def test_complicated_recipient_pagure_bad(
    rules,
    tahrir_client,
):
    rule = get_rule(rules, "Long Life to Pagure (Pagure I)")
    msg = Message(
        topic="io.pagure.prod.pagure.git.receive",
        body={
            "authors": [
                {
                    "fullname": "Pierre-YvesChibon",
                },
                {
                    "fullname": "Lubom\\u00edr Sedl\\u00e1\\u0159",
                },
            ],
            "total_commits": 2,
            "start_commit": "da090b8449237e3878d4d1fe56f7f8fcfd13a248",
        },
    )

    with (
        patch("fedbadges.rules.user_exists_in_fas") as g,
        patch("datanommer.models.Message.grep") as grep,
    ):
        grep.return_value = float("inf"), 1, MockQuery()
        g.return_value = True
        with pytest.raises(ValueError) as excinfo:
            rule.matches(msg, tahrir_client)
        assert str(excinfo.value) == "Multiple recipients: name not found in the message"



from unittest.mock import patch

import pytest
from fedora_messaging.message import Message

import fedbadges.rules

from .utils import example_real_bodhi_message


class MockQuery:
    def __init__(self, returned_count):
        self.returned_count = returned_count

    def count(self):
        return self.returned_count


def test_malformed_criteria():
    """Test that an error is raised when nonsense is provided."""
    with pytest.raises(KeyError):
        fedbadges.rules.Criteria(
            dict(
                watwat="does not exist",
            )
        )


def test_underspecified_criteria():
    """Test that an error is raised when condition is missing."""
    with pytest.raises(ValueError):
        fedbadges.rules.Criteria(
            dict(
                datanommer={
                    "filter": {"topics": ["%(topic)s"], "wat": "baz"},
                    "operation": "count",
                }
            )
        )


def test_malformed_filter():
    """Test that an error is raised for malformed filters"""
    with pytest.raises(KeyError):
        fedbadges.rules.Criteria(
            dict(
                datanommer={
                    "filter": {"topics": ["%(topic)s"], "wat": "baz"},
                    "operation": "count",
                    "condition": {
                        "greater than or equal to": 500,
                    },
                }
            )
        )


@pytest.mark.parametrize(
    ["returned_count", "expectation"],
    [
        (499, False),
        (500, True),
        (501, True),
    ],
)
def test_basic_datanommer(returned_count, expectation):
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "topics": ["%(topic)s"],
                },
                "operation": "count",
                "condition": {
                    "greater than or equal to": 500,
                },
            }
        )
    )
    message = Message(
        topic="org.fedoraproject.dev.something.sometopic",
    )
    with patch("datanommer.models.Message.grep") as grep:
        grep.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result == expectation
        grep.assert_called_once_with(
            topics=["org.fedoraproject.dev.something.sometopic"],
            defer=True,
        )


@pytest.mark.parametrize(
    ["returned_count", "expectation"],
    [
        (499, False),
        (500, True),
        (501, True),
    ],
)
def test_datanommer_with_lambda_condition(returned_count, expectation):
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "topics": ["%(topic)s"],
                },
                "operation": "count",
                "condition": {
                    "lambda": "value >= 500",
                },
            }
        )
    )
    message = Message(
        topic="org.fedoraproject.dev.something.sometopic",
    )
    with patch("datanommer.models.Message.grep") as f:
        f.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result == expectation


@pytest.mark.parametrize(
    ["returned_count", "expectation"],
    [
        (4, False),
        (5, True),
        (6, False),
    ],
)
def test_datanommer_formatted_operations(returned_count, expectation):
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "topics": ["%(topic)s"],
                },
                "operation": {
                    "lambda": "query.count() == %(msg.some_value)s",
                },
                "condition": {
                    "lambda": "value",
                },
            }
        )
    )
    message = Message(
        topic="org.fedoraproject.dev.something.sometopic",
        body=dict(
            some_value=5,
        ),
    )
    with patch("datanommer.models.Message.grep") as grep:
        grep.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result == expectation


@pytest.mark.parametrize(
    ["returned_count", "expectation"],
    [
        (499, False),
        (500, True),
        (501, True),
    ],
)
def test_datanommer_with_lambda_operation(returned_count, expectation):
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "topics": ["%(topic)s"],
                },
                "operation": {
                    "lambda": "query.count() - 5",
                },
                "condition": {
                    "lambda": "value >= 495",
                },
            }
        )
    )
    message = Message(
        topic="org.fedoraproject.dev.something.sometopic",
    )
    with patch("datanommer.models.Message.grep") as grep:
        grep.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result == expectation


def test_datanommer_with_lambda_filter():
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "users": {
                        "lambda": "[u for u in msg['message'].usernames "
                        "if not u in ['bodhi', 'hadess']]",
                    }
                },
                "operation": "count",
                "condition": {
                    "greater than or equal to": 0,
                },
            }
        )
    )

    message = example_real_bodhi_message
    returned_count = 0

    with patch("datanommer.models.Message.grep") as grep:
        grep.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result is True
        grep.assert_called_once_with(users=["lmacken"], defer=True)


def test_datanommer_with_dotted_filter():
    criteria = fedbadges.rules.Criteria(
        dict(
            datanommer={
                "filter": {
                    "users": [
                        "%(msg.update.user.name)s",
                    ]
                },
                "operation": "count",
                "condition": {
                    "greater than or equal to": 0,
                },
            }
        )
    )

    message = example_real_bodhi_message
    returned_count = 0

    with patch("datanommer.models.Message.grep") as grep:
        grep.return_value = returned_count, 1, MockQuery(returned_count)
        result = criteria.matches(message)
        assert result is True
        grep.assert_called_once_with(users=["hadess"], defer=True)



import logging

import click


_log = logging.getLogger(__name__)

option_debug = click.option("--debug", is_flag=True, default=False, help="enable debugg logging")


def setup_logging(debug=False):
    log_format = "[%(asctime)s][%(levelname)s] %(message)s"
    logging.basicConfig(format=log_format, level=logging.DEBUG if debug else logging.INFO)


def award_badge(tahrir, badge, email, check_existing=True):
    if check_existing and tahrir.assertion_exists(badge.id, email):
        _log.debug("%s already has %s, skipping", email, badge.id)
        return

    # time.sleep(1)
    _log.info("Awarding %s to %s", badge.id, email)
    try:
        tahrir.add_assertion(badge.id, email, None)
    except Exception:
        tahrir.session.rollback()
        _log.exception("Failure awarding badge %s to %s", badge.id, email)



""" Consumers for producing openbadges by listening for messages on fedmsg

Authors:  Ross Delinger
          Ralph Bean
          Aurelien Bompard
"""

import asyncio
import logging
import threading
import time
from functools import partial

import datanommer.models
import fasjson_client
import tahrir_api.dbapi
from fedora_messaging.api import Message
from fedora_messaging.config import conf as fm_config

from .aio import Periodic
from .rulesrepo import RulesRepo
from .utils import notification_callback


log = logging.getLogger(__name__)

DEFAULT_DELAY_LIMIT = 100
RULES_RELOAD_INTERVAL = 15 * 60  # 15 minutes


class FedoraBadgesConsumer:
    delay_limit = 100

    def __init__(self):
        self.config = fm_config["consumer_config"]
        self.delay_limit = int(self.config.get("delay_limit", DEFAULT_DELAY_LIMIT))
        self.badge_rules = []
        self.lock = threading.Lock()

        self.loop = asyncio.get_event_loop()
        self._ready = self.loop.create_task(self.setup())
        if not self.loop.is_running():
            self.loop.run_until_complete(self._ready)

    async def setup(self):
        # Five things need doing at start up time
        # 0) Set up a request local to hang thread-safe db sessions on.
        # 1) Initialize our connection to the Tahrir DB
        # 2) Initialize our connection to the datanommer DB.
        # 3) Load our badge definitions and rules from YAML.

        # Tahrir stuff.
        await self.loop.run_in_executor(None, self._initialize_tahrir_connection)

        # Datanommer stuff
        await self.loop.run_in_executor(None, self._initialize_datanommer_connection)

        # FASJSON stuff
        self.fasjson = await self.loop.run_in_executor(
            None, fasjson_client.Client, self.config["fasjson_base_url"]
        )

        # Load badge definitions
        self._rules_repo = RulesRepo(self.config, self.issuer_id, self.fasjson)
        self._rules_repo.setup()

        self._refresh_badges_task = Periodic(
            partial(self.loop.run_in_executor, None, self._reload_rules), RULES_RELOAD_INTERVAL
        )
        await self._refresh_badges_task.start(run_now=True)

    def _initialize_tahrir_connection(self):
        database_uri = self.config.get("database_uri")
        if not database_uri:
            raise ValueError("Badges consumer requires a database uri")
        issuer = self.config["badge_issuer"]
        self.tahrir = tahrir_api.dbapi.TahrirDatabase(
            dburi=database_uri,
            autocommit=False,
            notification_callback=notification_callback,
        )
        self.issuer_id = self.tahrir.add_issuer(
            issuer.get("issuer_origin"),
            issuer.get("issuer_name"),
            issuer.get("issuer_url"),
            issuer.get("issuer_email"),
        )
        self.tahrir.session.commit()

    def _get_tahrir_client(self, session=None):
        return self.tahrir

    def _initialize_datanommer_connection(self):
        datanommer.models.init(self.config["datanommer_db_uri"])

    def award_badge(self, username, badge_rule, link=None):
        email = f"{username}@fedoraproject.org"
        with self.TahrirDbSession() as session:
            client = self._get_tahrir_client(session)
            client.add_person(email)
            session.commit()
            client.add_assertion(badge_rule.badge_id, email, None, link)
            session.commit()

    def __call__(self, message: Message):
        # First thing, we receive the message, but we put ourselves to sleep to
        # wait for a moment.  The reason for this is that, when things are
        # 'calm' on the bus, we receive messages "too fast".  A message that
        # arrives to the badge awarder triggers (usually) a check against
        # datanommer to count messages.  But if we try to count them before
        # this message arrives at datanommer, we'll get skewed results!  Race
        # condition.  We go to sleep to allow ample time for datanommer to
        # consume this one before we go and start doing checks on it.
        # TODO: Option 1, easy: make a SQL query in datanommer instead of waiting
        # TODO: Option 2, a bit harder: check if the current message is in the
        #       matched messages when querying datanommer, and if it's not add 1
        #       to the count.

        self._wait_for_datanommer(message)

        datagrepper_url = self.config["datagrepper_url"]
        link = f"{datagrepper_url}/id?id={message.id}&is_raw=true&size=extra-large"

        # Define this so we can refer to it in error handling below
        badge_rule = None

        # Award every badge as appropriate.
        log.debug("Received %s, %s", message.topic, message.id)
        tahrir = self._get_tahrir_client()
        for badge_rule in self.badge_rules:
            try:
                for recipient in badge_rule.matches(message, tahrir):
                    log.debug(
                        "Rule %s matched for %s (message %s on %s)",
                        badge_rule.badge_id,
                        recipient,
                        message.id,
                        message.topic,
                    )
                    self.award_badge(recipient, badge_rule, link)
            except Exception:
                log.exception("Rule: %s, message: %s", repr(badge_rule), repr(message))

        log.debug("Done with %s, %s", message.topic, message.id)

    def _reload_rules(self):
        log.debug("Check for badges updates in the repo")
        tahrir = self._get_tahrir_client()
        self.badge_rules = self._rules_repo.load_all(tahrir)

    def _wait_for_datanommer(self, message: Message):
        while True:
            dn_msg = datanommer.models.Message.from_msg_id(message.id)
            if dn_msg is not None:
                break
            time.sleep(0.5)



def test_load_badges_number(consumer):
    """Determine that we can load badges from file."""
    assert len(consumer.badge_rules) == 5


def test_load_badges_contents(consumer):
    """Determine that we can load badges from file."""
    names = set([badge["name"] for badge in consumer.badge_rules])
    assert names == {
        "Like a Rock",
        "The Zen of Foo Bar Baz",
        "Junior Tagger (Tagger I)",
        "Speak Up!",
        "Long Life to Pagure (Pagure I)",
    }



import datetime
import logging

import click
import fasjson_client
from fedora_messaging.config import conf as fm_config
from tahrir_api.dbapi import TahrirDatabase

import fedbadges.utils

from .utils import award_badge, option_debug, setup_logging


# Users in the fedora-contributor group are members of at least 1 group.
REQUIRED_GROUPS = ["fedora-contributor"]

log = logging.getLogger(__name__)


def get_fas_userlist(fasjson, threshold):
    search_terms = dict(group=REQUIRED_GROUPS, creation__before=threshold)
    page_number = 0
    next_page_exists = True
    while next_page_exists:
        page_number += 1
        response = fasjson.search(
            page_size=40,
            page_number=page_number,
            _request_options={"headers": {"X-Fields": "username"}},
            **search_terms
        )
        yield from response.result
        next_page_exists = page_number < response.page["total_pages"]


@click.command()
@option_debug
def main(debug):
    setup_logging(debug=debug)
    config = fm_config["consumer_config"]
    uri = config["database_uri"]
    tahrir = TahrirDatabase(
        uri,
        notification_callback=fedbadges.utils.notification_callback,
    )
    badge = tahrir.get_badge(badge_id="badge-off!")
    if not badge:
        raise ValueError("badge does not exist")

    fasjson = fasjson_client.Client(config["fasjson_base_url"])

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    year = datetime.timedelta(days=365.5)
    mapping = {
        "egg": year * 1,
        "embryo": year * 2,
        "tadpole": year * 3,
        "tadpole-with-legs": year * 5,
        "froglet": year * 7,
        "adult-frog": year * 10,
    }

    # Query IPA for users created before the threshold
    for badge_id, delta in list(mapping.items()):
        badge = tahrir.get_badge(badge_id=badge_id)
        if not badge.id:
            log.error("Badge %s does not exist", badge_id)
            continue
        threshold = now - delta
        for person in get_fas_userlist(fasjson, threshold):
            email = person["username"] + "@fedoraproject.org"
            award_badge(tahrir, badge, email)


if __name__ == "__main__":
    main()



from fedora_messaging.message import Message

from fedbadges.utils import (
    construct_substitutions,
    format_args,
    single_argument_lambda_factory,
)


def test_lambda_factory():
    expression = "value + 2"
    target = 4
    actual = single_argument_lambda_factory(expression, 2)
    assert actual == target


def test_substitutions_basic():
    msg = Message(body=dict(a=dict(b=dict(c=42))))
    target = {
        "msg": dict(a=dict(b=dict(c=42))),
        "msg.a": dict(b=dict(c=42)),
        "msg.a.b": dict(c=42),
        "msg.a.b.c": 42,
        "topic": "",
    }
    actual = construct_substitutions(msg)
    assert actual == target


def test_substitutions_real():
    msg = Message(
        body={
            "thread": {"tagnames": ["town"], "pk": 2, "title": "alskdjflaksjdf lakjsf a"},
            "created": False,
            "timestamp": 1359947640.0,
            "topmost_post_id": 2,
            "agent": "ralph",
            "newly_mentioned_users": [],
            "diff": "<p>alskdfj... the diff is actually here",
            "post": {
                "vote_up_count": 0,
                "text": "alskdfjalskdjf alkjasdalskdjf ...",
                "summary": "alskdfjalskdjf alkjasdalskdjf ...",
                "comment_count": 0,
                "vote_down_count": 0,
                "pk": 2,
                "post_type": "question",
            },
        },
        topic="org.fedoraproject.dev.askbot.post.edit",
    )
    target = {
        "msg.post.text": "alskdfjalskdjf alkjasdalskdjf ...",
        "msg.thread.title": "alskdjflaksjdf lakjsf a",
        "msg.post.vote_down_count": 0,
        "msg.post.post_type": "question",
        "msg.thread.pk": 2,
        "msg.newly_mentioned_users": [],
        "msg.diff": "<p>alskdfj... the diff is actually here",
        "topic": "org.fedoraproject.dev.askbot.post.edit",
        "msg.agent": "ralph",
        "msg.post.comment_count": 0,
        "msg.post": {
            "vote_up_count": 0,
            "text": "alskdfjalskdjf alkjasdalskdjf ...",
            "summary": "alskdfjalskdjf alkjasdalskdjf ...",
            "comment_count": 0,
            "vote_down_count": 0,
            "pk": 2,
            "post_type": "question",
        },
        "msg.timestamp": 1359947640.0,
        "msg.topmost_post_id": 2,
        "msg.post.pk": 2,
        "msg.post.vote_up_count": 0,
        "msg.post.summary": "alskdfjalskdjf alkjasdalskdjf ...",
        "msg.thread.tagnames": ["town"],
        "msg.thread": {"tagnames": ["town"], "pk": 2, "title": "alskdjflaksjdf lakjsf a"},
        "msg": {
            "newly_mentioned_users": [],
            "thread": {"tagnames": ["town"], "pk": 2, "title": "alskdjflaksjdf lakjsf a"},
            "created": False,
            "topmost_post_id": 2,
            "timestamp": 1359947640.0,
            "post": {
                "vote_up_count": 0,
                "text": "alskdfjalskdjf alkjasdalskdjf ...",
                "summary": "alskdfjalskdjf alkjasdalskdjf ...",
                "comment_count": 0,
                "vote_down_count": 0,
                "pk": 2,
                "post_type": "question",
            },
            "diff": "<p>alskdfj... the diff is actually here",
            "agent": "ralph",
        },
        "msg.created": False,
    }
    actual = construct_substitutions(msg)
    assert actual == target


def test_format_args_simple():
    subs = {
        "foo.bar.baz": "value",
    }
    obj = {
        "something should be": "%(foo.bar.baz)s",
    }
    target = {
        "something should be": "value",
    }
    actual = format_args(obj, subs)
    assert actual == target


def test_format_args_list():
    subs = {
        "foo.bar.baz": "value",
    }
    obj = {
        "something should be": [
            "%(foo.bar.baz)s",
            "or this",
        ]
    }
    target = {
        "something should be": [
            "value",
            "or this",
        ]
    }
    actual = format_args(obj, subs)
    assert actual == target


def test_format_args_numeric():
    subs = {
        "foo.bar.baz": 42,
    }
    obj = {
        "something should be": "%(foo.bar.baz)i",
    }
    target = {
        "something should be": 42,
    }
    actual = format_args(obj, subs)
    assert actual == target


def test_format_args_nested():
    subs = {
        "wat": "another",
    }
    obj = {
        "one": {
            "thing": {
                "leads": {
                    "to": "%(wat)s",
                    "most": "of the time",
                }
            }
        }
    }
    target = {
        "one": {
            "thing": {
                "leads": {
                    "to": "another",
                    "most": "of the time",
                }
            }
        }
    }
    actual = format_args(obj, subs)
    assert actual == target


def test_format_args_nested_subs():
    subs = {"wat": dict(foo="bar")}
    obj = {
        "envelope": "%(wat)s",
    }
    target = {
        "envelope": dict(foo="bar"),
    }
    actual = format_args(obj, subs)
    assert actual == target



""" Models for fedbadges.

The primary thing here is a "BadgeRule" which is an in-memory working
abstraction of the trigger and criteria required to award a badge.

Authors:    Ralph Bean
"""

import abc
import copy
import functools
import inspect
import logging
import re

import datanommer.models
from fedora_messaging.api import Message
from tahrir_api.dbapi import TahrirDatabase

from fedbadges.utils import (
    # These are all in-process utilities
    construct_substitutions,
    email2fas,
    format_args,
    get_pagure_authors,
    graceful,
    nick2fas,
    recursive_lambda_factory,
    single_argument_lambda_factory,
    # These make networked API calls
    user_exists_in_fas,
)


log = logging.getLogger(__name__)


# Match OpenID agent strings, i.e. http://FAS.id.fedoraproject.org
def openid2fas(openid, config):
    id_provider_hostname = re.escape(config["id_provider_hostname"])
    m = re.search(f"^https?://([a-z][a-z0-9]+)\\.{id_provider_hostname}$", openid)
    if m:
        return m.group(1)
    return openid


def github2fas(uri, config, fasjson):
    m = re.search(r"^https?://api.github.com/users/([a-z][a-z0-9-]+)$", uri)
    if not m:
        log.warning("Can't extract the username from %r", uri)
        return None
    github_username = m.group(1)
    result = fasjson.search_users(github_username__exact=github_username)
    if len(result) != 1:
        return None
    return result[0]["username"]


def distgit2fas(uri, config):
    distgit_hostname = re.escape(config["distgit_hostname"])
    m = re.search(f"^https?://{distgit_hostname}/user/([a-z][a-z0-9]+)$", uri)
    if m:
        return m.group(1)
    return uri


def krb2fas(name):
    if "/" not in name:
        return name
    return name.split("/")[0]


def validate_possible(possible, fields):
    fields_set = set(fields)
    if not fields_set.issubset(possible):
        raise KeyError(
            f"{fields_set.difference(possible)!r} are not possible fields. "
            f"Choose from {possible!r}"
        )


def validate_badge(required, possible, badge_dict):
    fields = frozenset(list(badge_dict.keys()))
    validate_possible(possible, fields)
    if required and not required.issubset(fields):
        raise ValueError(
            f"BadgeRule requires {required!r}. Missing {required.difference(fields)!r}"
        )


operators = frozenset(
    [
        "all",
        "any",
        "not",
    ]
)
lambdas = frozenset(
    [
        "lambda",
    ]
)

operator_lookup = {"any": any, "all": all, "not": lambda x: all([not item for item in x])}


class BadgeRule:
    required = frozenset(
        [
            "name",
            "image_url",
            "description",
            "creator",
            "discussion",
            "issuer_id",
            "trigger",
            "criteria",
        ]
    )

    possible = required.union(
        [
            "recipient",
            "recipient_nick2fas",
            "recipient_email2fas",
            "recipient_openid2fas",
            "recipient_github2fas",
            "recipient_distgit2fas",
            "recipient_krb2fas",
        ]
    )

    banned_usernames = frozenset(
        [
            "bodhi",
            "oscar",
            "apache",
            "koji",
            "bodhi",
            "taskotron",
        ]
    )

    def __init__(self, badge_dict, issuer_id, config, fasjson):
        validate_badge(self.required, self.possible, badge_dict)
        self._d = badge_dict
        self.issuer_id = issuer_id
        self.config = config
        self.fasjson = fasjson

        self.trigger = Trigger(self._d["trigger"], self)
        self.criteria = Criteria(self._d["criteria"], self)
        self.recipient_key = self._d.get("recipient")
        self.recipient_nick2fas = self._d.get("recipient_nick2fas")
        self.recipient_email2fas = self._d.get("recipient_email2fas")
        self.recipient_openid2fas = self._d.get("recipient_openid2fas")
        self.recipient_github2fas = self._d.get("recipient_github2fas")
        self.recipient_distgit2fas = self._d.get("recipient_distgit2fas")
        self.recipient_krb2fas = self._d.get("recipient_krb2fas")

    def setup(self, tahrir: TahrirDatabase):
        self.badge_id = self._d["badge_id"] = tahrir.add_badge(
            name=self._d["name"],
            image=self._d["image_url"],
            desc=self._d["description"],
            criteria=self._d["discussion"],
            tags=",".join(self._d.get("tags", [])),
            issuer_id=self.issuer_id,
        )
        tahrir.session.commit()

    def __getitem__(self, key):
        return self._d[key]

    def __repr__(self):
        return f"<fedbadges.models.BadgeRule: {self._d!r}>"

    def matches(self, msg: Message, tahrir: TahrirDatabase):

        # First, do a lightweight check to see if the msg matches a pattern.
        if not self.trigger.matches(msg):
            return set()

        # Before proceeding further, let's see who would get this badge if
        # our more heavyweight checks matched up.  If the user specifies a
        # recipient_key, we can use that to extract the potential awardee.  If
        # that is not specified, we just use `msg.usernames`.
        if self.recipient_key:
            subs = construct_substitutions(msg)
            obj = format_args(self.recipient_key, subs)

            if isinstance(obj, (str, int, float)):
                obj = [obj]

            # On the way, it is possible for the fedmsg message to contain None
            # for "agent".  A problem here though is that None is not iterable,
            # so let's replace it with an equivalently empty iterable so code
            # further down doesn't freak out.  An instance of this is when a
            # user without a fas account comments on a bodhi update.
            if obj is None:
                obj = []

            # It is possible to recieve a list of dictionary containing the name
            # of the recipient, this is the case in the pagure's fedmsg.
            # In that case we create a new list containing the names taken from the
            # dictionnary.

            new_obj = get_pagure_authors(obj)
            if new_obj:
                obj = new_obj

            awardees = frozenset(obj)

            if self.recipient_nick2fas:
                awardees = frozenset([nick2fas(nick, self.fasjson) for nick in awardees])

            if self.recipient_email2fas:
                awardees = frozenset([email2fas(email, self.fasjson) for email in awardees])

            if self.recipient_openid2fas:
                awardees = frozenset([openid2fas(openid, self.config) for openid in awardees])

            if self.recipient_github2fas:
                awardees = frozenset(
                    [github2fas(uri, self.config, self.fasjson) for uri in awardees]
                )

            if self.recipient_distgit2fas:
                awardees = frozenset([distgit2fas(uri, self.config) for uri in awardees])

            if self.recipient_krb2fas:
                awardees = frozenset([krb2fas(uri) for uri in awardees])

            awardees = frozenset([e for e in awardees if e is not None])
        else:
            awardees = frozenset(msg.usernames)

        awardees = awardees.difference(self.banned_usernames)

        # Strip anyone who is an IP address
        awardees = frozenset(
            [
                user
                for user in awardees
                if not (user.startswith("192.168.") or user.startswith("10."))
            ]
        )

        # If no-one would get the badge by default, then no reason to waste
        # time doing any further checks.  No need to query the Tahrir DB.
        if not awardees:
            return awardees

        # Limit awardees to only those who do not already have this badge.
        awardees = frozenset(
            [
                user
                for user in awardees
                if not tahrir.assertion_exists(self.badge_id, f"{user}@fedoraproject.org")
                and not tahrir.person_opted_out(f"{user}@fedoraproject.org")
            ]
        )

        # If no-one would get the badge at this point, then no reason to waste
        # time doing any further checks.  No need to query datanommer.
        if not awardees:
            return awardees

        # Check our backend criteria -- likely, perform datanommer queries.
        try:
            if not self.criteria.matches(msg):
                return set()
        except OSError:
            log.exception("Failed checking criteria for rule %s", self.badge_id)
            return set()

        # Lastly, and this is probably most expensive.  Make sure the person
        # actually has a FAS account before we award anything.
        # https://github.com/fedora-infra/tahrir/issues/225
        awardees = set([u for u in awardees if user_exists_in_fas(self.fasjson, u)])

        return awardees


class AbstractComparator(metaclass=abc.ABCMeta):
    """Base class for shared behavior between trigger and criteria."""

    possible = required = frozenset()
    children = None

    def __init__(self, d, parent=None):
        validate_badge(self.required, self.possible, d)
        self._d = d
        self.parent = parent

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self._d!r}>, a child of {self.parent!r}"

    @abc.abstractmethod
    def matches(self, msg):
        pass


class AbstractTopLevelComparator(AbstractComparator):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls = type(self)

        if len(self._d) > 1:
            raise ValueError(
                "No more than one trigger allowed.  " "Use an operator, one of %r" % operators
            )
        self.attribute = next(iter(self._d))
        self.expected_value = self._d[self.attribute]

        # XXX - Check if we should we recursively nest Trigger/Criteria?

        # First, trick negation into thinking it is not a unary operator.
        if self.attribute == "not":
            self.expected_value = [self.expected_value]

        # Then, treat everything as if it accepts an arbitrary # of args.
        if self.attribute in operators:
            if not isinstance(self.expected_value, list):
                raise TypeError("Operators only accept lists, not %r" % type(self.expected_value))
            self.children = [cls(child, self) for child in self.expected_value]


class Trigger(AbstractTopLevelComparator):
    possible = (
        frozenset(
            [
                "topic",
                "category",
            ]
        )
        .union(operators)
        .union(lambdas)
    )

    @graceful(set())
    def matches(self, msg):
        # Check if we should just aggregate the results of our children.
        # Otherwise, we are a leaf-node doing a straightforward comparison.
        if self.children:
            return operator_lookup[self.attribute](child.matches(msg) for child in self.children)
        elif self.attribute == "lambda":
            return single_argument_lambda_factory(
                expression=self.expected_value,
                argument={"msg": msg.body, "message": msg},
                name="msg",
            )
        elif self.attribute == "category":
            return msg.topic.split(".")[3] == self.expected_value
        elif self.attribute == "topic":
            return msg.topic.endswith(self.expected_value)
        else:
            raise RuntimeError(f"Unexpected attribute: {self.attribute}")


class Criteria(AbstractTopLevelComparator):
    possible = frozenset(
        [
            "datanommer",
        ]
    ).union(operators)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not self.children:
            # Then, by AbstractComparator rules, I am a leaf node.  Specialize!
            self._specialize()

    def _specialize(self):
        if self.attribute == "datanommer":
            self.specialization = DatanommerCriteria(self.expected_value)
        # TODO -- expand this with other "backends" as necessary
        # elif self.attribute == 'fas'
        else:
            raise RuntimeError("This should be impossible to reach.")

    @graceful(set())
    def matches(self, msg):
        if self.children:
            return operator_lookup[self.attribute](child.matches(msg) for child in self.children)
        else:
            return self.specialization.matches(msg)


class AbstractSpecializedComparator(AbstractComparator):
    pass


class DatanommerCriteria(AbstractSpecializedComparator):
    required = possible = frozenset(
        [
            "filter",
            "operation",
            "condition",
        ]
    )

    condition_callbacks = {
        "is greater than or equal to": lambda t, v: v >= t,
        "greater than or equal to": lambda t, v: v >= t,
        "greater than": lambda t, v: v > t,
        "is less than or equal to": lambda t, v: v <= t,
        "less than or equal to": lambda t, v: v <= t,
        "less than": lambda t, v: v < t,
        "equal to": lambda t, v: v == t,
        "is equal to": lambda t, v: v == t,
        "is not": lambda t, v: v != t,
        "is not equal to": lambda t, v: v != t,
        "lambda": single_argument_lambda_factory,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if len(self._d["condition"]) > 1:
            conditions = list(self.condition_callbacks.keys())
            raise ValueError("No more than one condition allowed.  " "Use one of %r" % conditions)

        # Determine what arguments datanommer.models.Message.grep accepts
        argspec = inspect.getfullargspec(datanommer.models.Message.grep)
        irrelevant = frozenset(["defer"])
        grep_arguments = frozenset(argspec.args[1:]).difference(irrelevant)

        # Validate the filter
        validate_possible(grep_arguments, self._d["filter"])

        # Validate the condition
        condition_key, condition_val = next(iter(self._d["condition"].items()))
        if condition_key not in self.condition_callbacks:
            raise KeyError(
                f"{condition_key!r} is not a valid condition key. "
                f"Use one of {list(self.condition_callbacks)!r}"
            )

        # Construct a condition callable for later
        self.condition = functools.partial(self.condition_callbacks[condition_key], condition_val)

    def _construct_query(self, msg):
        """Construct a datanommer query for this message.

        The "filter" section of this criteria object will be used.  It will
        first be formatted with any substitutions present in the incoming
        message.  This is used, for instance, to construct a query like "give
        me all the messages bearing the same topic as the message that just
        arrived".
        """
        subs = construct_substitutions(msg)
        kwargs = format_args(copy.copy(self._d["filter"]), subs)
        kwargs = recursive_lambda_factory(kwargs, {"msg": msg.body, "message": msg}, name="msg")

        # It is possible to recieve a list of dictionary containing the name
        # of the recipient, this is the case in the pagure's fedmsg.
        # In that case we create a new list containing the names taken from the
        # dictionnary.
        if kwargs.get("users") is not None:
            for item in kwargs["users"]:
                if isinstance(item, list):
                    kwargs["users"] = item
            users = get_pagure_authors(kwargs["users"])
            if users:
                kwargs["users"] = users
        kwargs["defer"] = True
        total, pages, query = datanommer.models.Message.grep(**kwargs)
        return total, pages, query

    def _format_lambda_operation(self, msg):
        """Format the string representation of a lambda operation.

        The lambda operation can be formatted here to include strings that
        appear in the message being evaluated like
        %(msg.comment.update_submitter)s.  Placeholders like that will have
        their value substituted with whatever appears in the incoming message.
        """
        subs = construct_substitutions(msg)
        operation = format_args(copy.copy(self._d["operation"]), subs)
        return operation["lambda"]

    def matches(self, msg: Message):
        """A datanommer criteria check is composed of three steps.

        - A datanommer query is constructed by combining our yaml definition
          with the incoming fedmsg message that triggered us.
        - An operation in python is constructed by comining our yaml definition
          with the incoming fedmsg message that triggered us.  That operation
          is then executed against the datanommer query object.
        - A condition, derived from our yaml definition, is evaluated with the
          result of the operation from the previous step and is returned.
        """
        total, pages, query = self._construct_query(msg)
        if self._d["operation"] == "count":
            result = total
        elif isinstance(self._d["operation"], dict):
            expression = self._format_lambda_operation(msg)
            result = single_argument_lambda_factory(
                expression=expression, argument=query, name="query"
            )
        else:
            operation = getattr(query, self._d["operation"])
            result = operation()
        return self.condition(result)



""" Utilities for fedbadges that don't quite fit anywhere else. """

# These are here just so they're available in globals()
# for compiling lambda expressions
import json
import logging
import re
import sys
import traceback
import types

import backoff
from fedora_messaging import api as fm_api
from fedora_messaging import exceptions as fm_exceptions


log = logging.getLogger(__name__)


def construct_substitutions(message: fm_api.Message):
    subs = dict_to_subs({"msg": message.body})
    subs["topic"] = message.topic
    return subs


def dict_to_subs(msg: dict):
    """Convert a fedmsg message into a dict of substitutions."""
    subs = {}
    for key1 in msg:
        if isinstance(msg[key1], dict):
            subs.update(
                dict(
                    [
                        (".".join([key1, key2]), val2)
                        for key2, val2 in list(dict_to_subs(msg[key1]).items())
                    ]
                )
            )
            subs[key1] = msg[key1]
        elif isinstance(msg[key1], str):
            subs[key1] = msg[key1].lower()
        else:
            subs[key1] = msg[key1]
    return subs


def format_args(obj, subs):
    """Recursively apply a substitutions dict to a given criteria subtree"""

    if isinstance(obj, dict):
        for key in obj:
            obj[key] = format_args(obj[key], subs)
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return [format_args(item, subs) for item in obj]
    elif isinstance(obj, str) and obj[2:-2] in subs:
        obj = subs[obj[2:-2]]
    elif isinstance(obj, (int, float)):
        pass
    else:
        obj = obj % subs

    return obj


def single_argument_lambda_factory(expression, argument, name="value"):
    """Compile and execute a lambda expression with a single argument"""

    code = compile(f"lambda {name}: {expression}", __file__, "eval")
    lambda_globals = {
        "__builtins__": __builtins__,
        "json": json,
        "re": re,
    }
    func = types.LambdaType(code, lambda_globals)()
    return func(argument)


def recursive_lambda_factory(obj, arg, name="value"):
    """Given a dict, find any lambdas, compile, and execute them."""

    if isinstance(obj, dict):
        for key in obj:
            if key == "lambda":
                # If so, *replace* the parent dict with the result of the expr
                obj = single_argument_lambda_factory(obj[key], arg, name)
                break
            else:
                obj[key] = recursive_lambda_factory(obj[key], arg, name)
    elif isinstance(obj, list):
        return [recursive_lambda_factory(e, arg, name) for e in obj]
    else:
        pass

    return obj


def graceful(default_return_value):
    """A decorator that gracefully handles exceptions."""

    def decorate(method):
        def inner(self, *args, **kwargs):
            try:
                return method(self, *args, **kwargs)
            except Exception:
                log.exception(
                    "From method: %r self: %r args: %r kwargs: %r", method, self, args, kwargs
                )
                return default_return_value

        return inner

    return decorate


def _backoff_hdlr(details):
    log.warning(f"Publishing message failed. Retrying. {traceback.format_tb(sys.exc_info()[2])}")


@backoff.on_exception(
    backoff.expo,
    (fm_exceptions.ConnectionException, fm_exceptions.PublishException),
    max_tries=3,
    on_backoff=_backoff_hdlr,
)
def _publish(message):
    fm_api.publish(message)


def notification_callback(message):
    """This is a callback called by tahrir_api whenever something
    it deems important has happened.

    It is just used to publish fedmsg messages.
    """
    try:
        _publish(message)
    except fm_exceptions.BaseException:
        log.error(f"Publishing message failed. Giving up. {traceback.format_tb(sys.exc_info()[2])}")


def user_exists_in_fas(fasjson, user):
    """Return true if the user exists in FAS."""
    return fasjson.get_user(username=user) is not None


def get_pagure_authors(authors):
    """Extract the name of pagure authors from
    a dictionary

    Args:
    authors (list): A list of dict that contains fullname and name key.
    """
    authors_name = []
    for item in authors:
        if isinstance(item, dict):
            try:
                if item["name"] is not None:
                    authors_name.append(item["name"])
            except KeyError as e:
                raise ValueError("Multiple recipients: name not found in the message") from e
    return authors_name


def nick2fas(nick, fasjson):
    """Return the user in FAS."""
    return fasjson.get_user(username=nick)


def email2fas(email, fasjson):
    """Return the user with the specified email in FAS."""
    if email.endswith("@fedoraproject.org"):
        return nick2fas(email.rsplit("@", 1)[0], fasjson)

    result = fasjson.search_users(email=email)
    if not result:
        return None
    return result[0]



import logging

import click
import requests
from fedora_messaging.config import conf as fm_config
from tahrir_api.dbapi import TahrirDatabase

import fedbadges.utils

from .utils import award_badge, option_debug, setup_logging


log = logging.getLogger(__name__)

HTTP_TIMEOUT = 5


@click.command()
@option_debug
def main(debug):
    setup_logging(debug=debug)
    config = fm_config["consumer_config"]
    uri = config["database_uri"]
    tahrir = TahrirDatabase(
        uri,
        notification_callback=fedbadges.utils.notification_callback,
    )

    url = "https://admin.fedoraproject.org/mirrormanager/api/mirroradmins"
    response = requests.get(url, timeout=HTTP_TIMEOUT)

    if not response.ok:
        raise click.ClickException("Couldn't read the mirrormanager/mirroradmins list.")

    usernames = response.json()["admins"]

    badge = tahrir.get_badge(badge_id="mirror,-mirror-on-the-wall")
    if not badge:
        raise ValueError("Badge does not exist")

    for username in usernames:
        email = f"{username}@fedoraproject.org"

        person = tahrir.get_person(person_email=email)
        if not person:
            log.info("%s does not exist.  Creating.", email)
            tahrir.add_person(email, nickname=username)

        award_badge(tahrir, badge, email)


if __name__ == "__main__":
    main()



import logging
from collections import defaultdict

import click
import fasjson_client
from fedora_messaging.config import conf as fm_config
from tahrir_api.dbapi import TahrirDatabase

import fedbadges.utils

from .utils import award_badge, option_debug, setup_logging


log = logging.getLogger(__name__)


def get_awards(fasjson, group_badges):
    awards = defaultdict(list)
    for group_badge in group_badges:
        group_name = group_badge["group"]
        badge_id = group_badge["badge"]
        membership_types = group_badge.get("memberships", ("member", "sponsor"))
        log.debug("Processing %s", group_name)
        for membership_type in membership_types:
            method = getattr(fasjson, f"list_group_{membership_type}s")
            try:
                members = method(groupname=group_name).result
            except fasjson_client.errors.APIError:
                continue
            for user in members:
                username = user["username"]
                awards[badge_id].append(username)

    return awards


@click.command()
@option_debug
def main(debug):
    setup_logging(debug=debug)
    config = fm_config["consumer_config"]
    uri = config["database_uri"]
    tahrir = TahrirDatabase(
        uri,
        notification_callback=fedbadges.utils.notification_callback,
    )
    fasjson = fasjson_client.Client(config["fasjson_base_url"])

    group_badges = config["group_badges"]

    # First, some validation that the badge ids actually exist.
    for badge_id in [gb["badge"] for gb in group_badges]:
        badge = tahrir.get_badge(badge_id=badge_id)
        if not badge:
            raise click.ClickException(f"{badge_id!r} is not a valid badge id")

    # Then, do a long query against FAS for our candidates.
    awards = get_awards(fasjson, group_badges)

    for badge_id, usernames in awards.items():
        badge = tahrir.get_badge(badge_id=badge_id)
        for username in usernames:
            award_badge(tahrir, badge, username)



import datetime
import logging
import os
import subprocess

import yaml

import fedbadges.rules


log = logging.getLogger(__name__)


class RulesRepo:

    def __init__(self, config, issuer_id, fasjson):
        self.config = config
        self.issuer_id = issuer_id
        self.fasjson = fasjson
        self.directory = os.path.abspath(self.config["badges_repo"])
        self._last_rules_load = None
        self.rules = []

    def setup(self):
        self._mark_safe()

    def _mark_safe(self):
        result = subprocess.run(
            ["/usr/bin/git", "config", "--get-all", "safe.directory"],  # noqa: S603
            text=True,
            stdout=subprocess.PIPE,
        )
        if result.returncode == 1:
            # Option isn't set
            safe_dirs = []
        else:
            result.check_returncode()
            safe_dirs = result.stdout.strip().split("\n")
        log.debug("Git safe dirs: %r", safe_dirs)
        if self.directory not in safe_dirs:
            log.debug("Adding %s to git's safe dirs", self.directory)
            subprocess.run(
                [  # noqa: S603
                    "/usr/bin/git",
                    "config",
                    "--global",
                    "--add",
                    "safe.directory",
                    self.directory,
                ],
                check=True,
            )

    def load_all(self, tahrir_client, force=False):
        if force or self._needs_update():
            self.rules = self._load_all(tahrir_client)
        return self.rules

    def _load_all(self, tahrir_client):
        self._last_rules_load = datetime.datetime.now(tz=datetime.timezone.utc)
        rules_dir = os.path.join(self.directory, "rules")
        # badges indexed by trigger
        badges = []
        log.info("Looking in %r to load badge definitions", rules_dir)
        for root, _dirs, files in os.walk(rules_dir):
            for partial_fname in files:
                fname = root + "/" + partial_fname
                badge = self._load_badge_from_yaml(fname)

                if not badge:
                    continue

                try:
                    badge_rule = fedbadges.rules.BadgeRule(
                        badge, self.issuer_id, self.config, self.fasjson
                    )
                    badge_rule.setup(tahrir_client)
                    badges.append(badge_rule)
                except ValueError as e:
                    log.error("Initializing rule for %r failed with %r", fname, e)

        log.info("Loaded %s total badge definitions", len(badges))
        return badges

    def _load_badge_from_yaml(self, fname):
        log.debug("Loading %r" % fname)
        try:
            with open(fname) as f:
                return yaml.safe_load(f.read())
        except Exception as e:
            log.error("Loading %r failed with %r", fname, e)
            return None

    def _needs_update(self):
        if self._last_rules_load is None:
            return True
        result = subprocess.run(
            [  # noqa: S603
                "/usr/bin/git",
                "-C",
                self.directory,
                "log",
                "-1",
                r"--pretty=format:%aI",
            ],
            check=True,
            stdout=subprocess.PIPE,
            text=True,
        )
        last_commit_time = datetime.datetime.fromisoformat(result.stdout.strip())
        if last_commit_time.tzinfo is None:
            last_commit_time = last_commit_time.replace(tzinfo=datetime.timezone.utc)
        return last_commit_time > self._last_rules_load



import asyncio
import inspect
import logging
from contextlib import suppress


log = logging.getLogger(__name__)


class Periodic:
    """
    Inspired by https://stackoverflow.com/a/37514633

    This runs a task at intervals of the provided seconds.
    The first run will happen after the first interval, unless `run_now` is set to True in the
    `start()` call.
    For long-running tasks, note that there will be `seconds` between two tasks regardless of when
    the task started.
    """

    def __init__(self, func, seconds: float):
        self.func = func
        self.seconds = seconds
        self.is_started = False
        self._task = None

    async def start(self, run_now=False):
        if not self.is_started:
            self.is_started = True
            if run_now:
                await self._run_func()
            # Start task to call func periodically:
            self._task = asyncio.create_task(self._run_loop())

    async def stop(self):
        if self.is_started:
            self.is_started = False
            # Stop task and await it stopped:
            self._task.cancel()
            with suppress(asyncio.CancelledError):
                await self._task

    async def _run_loop(self):
        while True:
            await asyncio.sleep(self.seconds)
            await self._run_func()

    async def _run_func(self):
        try:
            result = self.func()
            if inspect.isawaitable(result):
                await result
        except Exception:
            log.exception("Periodic call failed:")



from unittest.mock import Mock, patch

import pytest
from fedora_messaging.message import Message
from noggin_messages import MemberSponsorV1

import fedbadges.rules

from .utils import example_real_bodhi_message


class MockQuery:
    def count():
        return 1


def test_metadata_validation():
    """Test for failure if not enough metadata"""
    with pytest.raises(ValueError):
        fedbadges.rules.BadgeRule(dict(name="Test"), None, None, None)


def test_full_specification():
    """Test for failure if no criteria specified."""
    with pytest.raises(ValueError):
        fedbadges.rules.BadgeRule(
            dict(
                name="Test",
                description="Doesn't matter...",
                creator="Somebody",
                discussion="http://somelink.com",
                issuer_id="fedora-project",
                image_url="http://somelinke.com/something.png",
                trigger=dict(topic="test_topic"),
            ),
            1,
            None,
            None,
        )


def test_full_simple_success(fasjson_client, tahrir_client):
    """A simple integration test for messages with zero users"""
    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="bodhi"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    msg = example_real_bodhi_message

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        with patch("fedbadges.rules.user_exists_in_fas") as g:
            g.return_value = True
            assert rule.matches(msg, tahrir_client) == set(["lmacken", "hadess"])


def test_full_simple_match_almost_succeed(fasjson_client, tahrir_client):
    """A simple integration test for messages with zero users"""
    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="bodhi"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    # This message has zero users associated with it, so even if trigger
    # and criteria are met, there's noone to award the badge to -- and so,
    # we should *fail* the ``matches`` call.
    msg = Message(topic="org.fedoraproject.prod.bodhi.mashtask.complete", body={"success": False})

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = None, None, MockQuery()
        assert rule.matches(msg, tahrir_client) == set()


def test_yaml_specified_awardee_success(fasjson_client, tahrir_client):
    """Test that we can override msg.usernames."""
    # For instance, fas.group.member.remove contains two users,
    # the one being removed from a group, and the one doing the removing.
    # a badge YAML definition needs to be able to specify *which* of these
    # two users should receive its badge.  The dotted substitution
    # notation should suffice for this.  If this is specified, use it.
    # If not, use msg.usernames for convenience.  It will do
    # in most all cases.

    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="fas"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
            recipient="%(msg.agent.username)s",
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    msg = Message(
        topic="org.fedoraproject.stg.fas.role.update",
        body={
            "group": {"name": "ambassadors"},
            "user": {"username": "ralph"},
            "agent": {"username": "toshio"},
        },
    )

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        with patch("fedbadges.rules.user_exists_in_fas") as g:
            g.return_value = True
            assert rule.matches(msg, tahrir_client) == set(["toshio"])


def test_yaml_specified_awardee_failure(fasjson_client, tahrir_client):
    """Test that when we don't override msg.usernames, we get 2 awardees."""
    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="fas"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    msg = MemberSponsorV1(
        topic="org.fedoraproject.stg.fas.group.member.sponsor",
        body={
            "msg": {
                "group": "ambassadors",
                "user": "ralph",
                "agent": "toshio",
            }
        },
    )

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        with patch("fedbadges.rules.user_exists_in_fas") as g:
            g.return_value = True
            assert rule.matches(msg, tahrir_client) == set(["toshio", "ralph"])


def test_against_duplicates(fasjson_client, tahrir_client):
    """Test that matching fails if user already has the badge."""

    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="fas"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)
    tahrir_client.add_person("toshio@fedoraproject.org")
    tahrir_client.session.commit()
    tahrir_client.add_assertion(rule.badge_id, "toshio@fedoraproject.org", None, None)
    tahrir_client.session.commit()
    print(tahrir_client.assertion_exists(rule.badge_id, "toshio@fedoraproject.org"))
    print(rule.badge_id)

    msg = MemberSponsorV1(
        topic="org.fedoraproject.stg.fas.group.member.sponsor",
        body={
            "msg": {
                "group": "ambassadors",
                "user": "ralph",
                "agent": "toshio",
            }
        },
    )

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        with patch("fedbadges.rules.user_exists_in_fas") as g:
            g.return_value = True
            assert rule.matches(msg, tahrir_client) == set(["ralph"])


@patch("fedbadges.rules.user_exists_in_fas", Mock(return_value=True))
def test_github_awardee(fasjson_client, tahrir_client):
    """Conversion from GitHub URI to FAS users"""
    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="bodhi"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
            recipient="%(msg.user)s",
            recipient_github2fas="Yes",
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    msg = Message(
        topic="org.fedoraproject.prod.bodhi.update.request.testing",
        body={"user": "https://api.github.com/users/dummygh"},
    )

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        fasjson_client.search_users.return_value = [{"username": "dummy"}]
        assert rule.matches(msg, tahrir_client) == set(["dummy"])
        fasjson_client.search_users.assert_called_once_with(github_username__exact="dummygh")


def test_krb_awardee(fasjson_client, tahrir_client):
    """Conversion from Kerberos user to FAS users"""
    rule = fedbadges.rules.BadgeRule(
        dict(
            name="Test",
            description="Doesn't matter...",
            creator="Somebody",
            discussion="http://somelink.com",
            issuer_id="fedora-project",
            image_url="http://somelinke.com/something.png",
            trigger=dict(category="buildsys"),
            criteria=dict(
                datanommer=dict(
                    filter=dict(categories=["pkgdb"]),
                    operation="count",
                    condition={"greater than or equal to": 1},
                )
            ),
            recipient="%(msg.owner)s",
            recipient_krb2fas="Yes",
        ),
        1,
        None,
        fasjson_client,
    )
    rule.setup(tahrir_client)

    # msg = Message(
    #     id="cedd7ab4-8a59-4704-bd1b-0e7297bf759c",
    #     topic="org.fedoraproject.prod.buildsys.build.state.change",
    #     headers={
    #         "sent-at": "2022-06-29T16:27:05+00:00",
    #     },
    #     body={
    #         "build_id": 1994993,
    #         "old": 0,
    #         "name": "dummy-test-package-gloster",
    #         "task_id": 88890394,
    #         "instance": "primary",
    #         "attribute": "state",
    #         "request": [
    #             "git+https://src.fedoraproject.org/rpms/dummy-test-package-gloster.git#aaf707bc5671ab5e00c5618d95bfd83803ca54c0",
    #             "rawhide",
    #             {},
    #         ],
    #         "owner": "packagerbot/os-master02.iad2.fedoraproject.org",
    #         "epoch": None,
    #         "version": "0",
    #         "release": "9242.fc37",
    #         "new": 1,
    #     },
    # )
    msg = Message(
        topic="org.fedoraproject.prod.buildsys.build.state.change",
        body={"owner": "packagerbot/os-master02.iad2.fedoraproject.org"},
    )

    with patch("datanommer.models.Message.grep") as f:
        f.return_value = 1, 1, MockQuery()
        with patch("fedbadges.rules.user_exists_in_fas") as g:
            g.return_value = True
            assert rule.matches(msg, tahrir_client) == set(["packagerbot"])



import logging
import subprocess
import tempfile

import click
import fasjson_client
from fedora_messaging.config import conf as fm_config
from tahrir_api.dbapi import TahrirDatabase

import fedbadges.utils

from .utils import award_badge, option_debug, setup_logging


log = logging.getLogger(__name__)


badge = None
http_client = None


def gather_authors(repodir):
    result = subprocess.run(
        ["/usr/bin/git", "log", r"--pretty=tformat:%ae"],  # noqa: S603
        cwd=repodir,
        text=True,
        stdout=subprocess.PIPE,
    )
    yield from set(result.stdout.splitlines())


def email_to_fas_accounts(fasjson, email):
    response = fasjson.search(email=email)
    if response.page["total_results"] != 1:
        return None
    return response.result[0]["username"]


@click.command()
@option_debug
def main(debug):
    setup_logging(debug=debug)
    config = fm_config["consumer_config"]
    uri = config["database_uri"]
    tahrir = TahrirDatabase(
        uri,
        notification_callback=fedbadges.utils.notification_callback,
    )
    badge = tahrir.get_badge(badge_id="badge-off!")
    if not badge:
        raise ValueError("badge does not exist")
    fasjson = fasjson_client.Client(config["fasjson_base_url"])
    repos = ["tahrir", "tahrir-api", "fedbadges"]
    for repo in repos:
        with tempfile.TemporaryDirectory() as tmpdir:
            log.info("Trying %s", repo)
            subprocess.run(
                [  # noqa: S603
                    "/usr/bin/git",
                    "clone",
                    f"https://github.com/fedora-infra/{repo}.git",
                    tmpdir,
                ]
            )
            for email in gather_authors(tmpdir):
                log.debug("Considering email %s", email)
                username = email_to_fas_accounts(fasjson, email)
                log.debug("Considering user %s", username)
                award_badge(tahrir, badge, username)


if __name__ == "__main__":
    main()



import pytest
from fedora_messaging.message import Message

import fedbadges.rules


def test_basic_topic_matching_isolated():
    """Test that the matches method can match a basic topic."""
    trigger = fedbadges.rules.Trigger(
        dict(
            topic="test_topic",
        )
    )
    message = Message(
        topic="test_topic",
    )
    assert trigger.matches(message)


def test_basic_category_matching_isolated():
    """Test that the matches method can match a basic category."""
    trigger = fedbadges.rules.Trigger(
        dict(
            category="test_category",
        )
    )
    message = Message(
        topic="org.fedoraproject.dev.test_category.some_topic",
    )
    assert trigger.matches(message)


def test_basic_conjunction_pass():
    """Test that two anded fields accept the right message"""
    trigger = fedbadges.rules.Trigger(
        {
            "all": [
                dict(topic="org.fedoraproject.dev.test_category.test_topic"),
                dict(category="test_category"),
            ]
        }
    )
    message = Message(
        topic="org.fedoraproject.dev.test_category.test_topic",
    )
    assert trigger.matches(message)


def test_basic_conjunction_fail():
    """Test that two anded fields reject non-matching messages"""
    trigger = fedbadges.rules.Trigger(
        {
            "all": [
                dict(topic="org.fedoraproject.dev.test_category.test_topic"),
                dict(category="test_category"),
            ]
        }
    )
    message = Message(
        topic="org.fedoraproject.dev.test_category.test_topic.doesntmatch",
    )
    assert not trigger.matches(message)


def test_lambdas_pass():
    """Test that lambdas match correctly"""
    trigger = fedbadges.rules.Trigger(
        {
            "lambda": "'s3kr3t' in json.dumps(msg['msg'])",
        }
    )
    message = Message(body=dict(nested=dict(something="s3kr3t")))
    assert trigger.matches(message)


def test_lambdas_fail():
    """Test that lambdas fail correctly"""
    trigger = fedbadges.rules.Trigger(
        {
            "lambda": "'one string' in json.dumps(msg['msg'])",
        }
    )
    message = Message(body=dict(nested=dict(something="another string")))
    assert not trigger.matches(message)


def test_invalid_nesting():
    """Test that invalid nesting is detected and excepted."""
    with pytest.raises(TypeError):
        fedbadges.rules.Trigger(
            {
                "all": dict(
                    topic="org.fedoraproject.dev.test_category.test_topic",
                    category="test_category",
                )
            }
        )


def test_two_fields():
    """Test that passing two statements as a trigger is invalid."""
    with pytest.raises(ValueError):
        fedbadges.rules.Trigger(
            dict(
                topic="test_topic",
                category="test_topic",
            )
        )


def test_malformed_trigger():
    """Test that a single, undefined field is handled as invalid."""
    with pytest.raises(KeyError):
        fedbadges.rules.Trigger(
            dict(
                watwat="does not exist",
            )
        )



import hashlib
import logging

import click
import requests
from fedora_messaging.config import conf as fm_config
from tahrir_api.dbapi import TahrirDatabase

import fedbadges.utils

from .utils import award_badge, option_debug, setup_logging


log = logging.getLogger(__name__)

HTTP_TIMEOUT = 5


@click.command()
@option_debug
def main(debug):
    setup_logging(debug=debug)
    config = fm_config["consumer_config"]
    uri = config["database_uri"]
    tahrir = TahrirDatabase(
        uri,
        notification_callback=fedbadges.utils.notification_callback,
    )
    badge = tahrir.get_badge("mugshot")

    persons = tahrir.get_all_persons()
    already_has_it = [assertion.person for assertion in badge.assertions]

    good, bad = [], []
    for person in persons:

        if person in already_has_it:
            good.append(person)
            log.debug("Skipping %s", person)
            continue

        openid = f"http://{person.nickname}.id.fedoraproject.org/"
        hash = hashlib.sha256(openid.encode("utf-8")).hexdigest()
        url = f"https://seccdn.libravatar.org/avatar/{hash}?d=404"

        response = None
        for i in range(10):
            log.debug("Try %s on %s", i, url)
            try:
                response = requests.get(url, timeout=HTTP_TIMEOUT)
                break
            except requests.exceptions.SSLError as e:
                print(" * failed, trying again", str(e))

        if response is None:
            raise

        if response.status_code == 200:
            log.info("%s totally gets the mugshot badge.", person.nickname)
            good.append(person)
            award_badge(tahrir, badge.id, person.email, check_existing=False)
        else:
            bad.append(person)

    log.info("%s good peoples", len(good))
    log.info("%s bad peoples", len(bad))


if __name__ == "__main__":
    main()
